"""Elements that hang off the hub."""
import aiopulse.utils as utils
import aiopulse.const as const


class Roller:
    """Representation of a Roller blind."""

    def __init__(self, hub, roller_id):
        """Init a new roller blind."""
        self.hub = hub
        self.id = roller_id
        self.name = None
        self.serial = None
        self.room_id = None
        self.room = None
        self.battery = None
        self.closed_percent = None
        self.flags = 0
        self.update_callback = None

    def __str__(self):
        """Returns string representation of roller."""
        return (
            "Name: {} ID: {} Serial: {} Room: {} Closed %: {} Battery %: {}"
            " Flags: {:08b}"
        ).format(
            self.name,
            self.id,
            self.serial,
            self.room.name,
            self.closed_percent,
            self.battery,
            self.flags,
        )

    def set_callback(self, callback):
        """Add a callback for device updates."""
        self.update_callback = callback

    def notify_callback(self):
        """Tell callback that device has been updated."""
        if self.update_callback:
            self.update_callback()

    async def move_to(self, percent):
        """Send command to move the roller to a percentage closed."""
        message = (
            bytes.fromhex("0000000000000101")
            + bytes.fromhex("0600")
            + utils.pack_int(self.id, 6)
            + bytes.fromhex("03010100")
            + bytes.fromhex("190401030001")
            + utils.pack_int(percent, 2)
            + bytes.fromhex("ff")
        )
        await self.hub.send_payload(
            const.COMMAND_MOVE_TO, bytes.fromhex("2201"), message
        )

    async def move_up(self):
        """Send command to move the roller to fully open."""
        message = (
            bytes.fromhex("0000000000000101")
            + bytes.fromhex("0600")
            + utils.pack_int(self.id, 6)
            + bytes.fromhex("03010100")
            + bytes.fromhex("10")
            + bytes.fromhex("ff")
        )
        await self.hub.send_payload(const.COMMAND_MOVE, bytes.fromhex("2201"), message)

    async def move_stop(self):
        """Send command to stop the roller."""
        message = (
            bytes.fromhex("0000000000000101")
            + bytes.fromhex("0600")
            + utils.pack_int(self.id, 6)
            + bytes.fromhex("03010100")
            + bytes.fromhex("11")
            + bytes.fromhex("ff")
        )
        await self.hub.send_payload(const.COMMAND_MOVE, bytes.fromhex("2201"), message)

    async def move_down(self):
        """Send command to move the roller to fully closed."""
        message = (
            bytes.fromhex("0000000000000101")
            + bytes.fromhex("0600")
            + utils.pack_int(self.id, 6)
            + bytes.fromhex("03010100")
            + bytes.fromhex("12")
            + bytes.fromhex("ff")
        )
        await self.hub.send_payload(const.COMMAND_MOVE, bytes.fromhex("2201"), message)


class Room:
    """Representation of a Room."""

    def __init__(self, hub, room_id):
        """Init a new room."""
        self.hub = hub
        self.id = room_id
        self.icon = None
        self.name = None
        self.update_callback = None

    def __str__(self):
        """Returns string representation of roller."""
        return "Name: {} ID: {} Icon: {}".format(self.name, self.id[0:4], self.icon)

    def set_callback(self, callback):
        """Add a callback for device updates."""
        self.update_callback = callback

    def notify_callback(self):
        """Tell callback that device has been updated."""
        if self.update_callback:
            self.update_callback()

    async def move_to(self, percent):
        """Send command to move the roller to a percentage closed."""
        message = (
            bytes.fromhex("0000000000000101")
            + bytes.fromhex("0600")
            + utils.pack_int(self.id, 6)
            + bytes.fromhex("03010100")
            + bytes.fromhex("190401030001")
            + utils.pack_int(percent, 2)
            + bytes.fromhex("ff")
        )
        await self.hub.send_payload(
            const.COMMAND_MOVE_TO, bytes.fromhex("2201"), message
        )

    async def move_up(self):
        """Send command to move the roller to fully open."""
        message = (
            bytes.fromhex("0000000000000101")
            + bytes.fromhex("0600")
            + utils.pack_int(self.id, 6)
            + bytes.fromhex("03010100")
            + bytes.fromhex("10")
            + bytes.fromhex("ff")
        )
        await self.hub.send_payload(const.COMMAND_MOVE, bytes.fromhex("2201"), message)

    async def move_stop(self):
        """Send command to stop the roller."""
        message = (
            bytes.fromhex("0000000000000101")
            + bytes.fromhex("0600")
            + utils.pack_int(self.id, 6)
            + bytes.fromhex("03010100")
            + bytes.fromhex("11")
            + bytes.fromhex("ff")
        )
        await self.hub.send_payload(const.COMMAND_MOVE, bytes.fromhex("2201"), message)

    async def move_down(self):
        """Send command to move the roller to fully closed."""
        message = (
            bytes.fromhex("0000000000000101")
            + bytes.fromhex("0600")
            + utils.pack_int(self.id, 6)
            + bytes.fromhex("03010100")
            + bytes.fromhex("12")
            + bytes.fromhex("ff")
        )
        await self.hub.send_payload(const.COMMAND_MOVE, bytes.fromhex("2201"), message)


class Scene:
    """Representation of a Scene."""

    def __init__(self, hub, scene_id):
        """Init a new scene."""
        self.hub = hub
        self.id = scene_id
        self.icon = None
        self.name = None

    def __str__(self):
        """Returns string representation of roller."""
        return "Name: {} ID: {} Icon: {}".format(self.name, self.id[0:4], self.icon)



""" Serialisation / Deserialisation helpers """

def unpack_int(buffer, ptr, length):
	""" Unpack an int of specified length from the buffer and advance the pointer """
	return (
		int.from_bytes(buffer[ptr:(ptr+length)], 'little', signed=False),
		ptr + length
		)

def pack_int(value, length):
	""" Unpack an int for serialisation """
	return value.to_bytes(length, 'little', signed=False)

def unpack_bytes(buffer, ptr, length=None):
	""" Unpack a specified number of bytes from the buffer and advance the pointer """
	ptr_new = ptr
	if not length:
		length, ptr_new = unpack_int(buffer, ptr, 2)
	return (
		buffer[(ptr_new):(ptr_new+length)],
		ptr_new + length
		)
		
def unpack_string(buffer, ptr, length=None):
	""" Unpack a specified number of characters from the buffer and advance the pointer """
	str_new, ptr_new = unpack_bytes(buffer, ptr, length=None)
	return (
		str_new.decode("utf-8"),
		ptr_new
		)


class HubBaseException(Exception):
	""" Base Exception for protocol """
	pass

class NotConnectedException(HubBaseException):
	""" Exception thrown when the hub isn't connected """
	pass

class NotRunningException(HubBaseException):
	""" Exception thrown when the hub isn't running """
	pass

class InvalidResponseException(HubBaseException):
	""" Exception thrown when an invalid response is received """
	pass


"""Rollease Acmeda Automate Pulse asyncio protocol implementation."""
import logging

from aiopulse.hub import Hub
from aiopulse.elements import Roller, Room, Scene
from aiopulse.errors import (
    NotConnectedException,
    NotRunningException,
    InvalidResponseException,
)

__all__ = [
    "Hub",
    "Roller",
    "Room",
    "Scene",
    "NotConnectedException",
    "NotRunningException",
    "InvalidResponseException",
]
__version__ = "0.2.2
__author__ = "Alan Murray"

_LOGGER = logging.getLogger(__name__)




HEADER = 				bytes.fromhex('00000003')
COMMAND_DISCOVER = 		bytes.fromhex('03000003')
RESPONSE_DISCOVER = 	bytes.fromhex('57000004')

COMMAND_CONNECT = 		bytes.fromhex('03000006')
RESPONSE_CONNECT = 		bytes.fromhex('0f000007')

COMMAND_LOGIN = 		bytes.fromhex('0f000008')
RESPONSE_LOGIN = 		bytes.fromhex('04000009')

COMMAND_PING = 			bytes.fromhex('03000015')
RESPONSE_PING = 		bytes.fromhex('03000016')

COMMAND_SETID = 		bytes.fromhex('28000090')
RESPONSE_SETID = 		bytes.fromhex('03000091')

COMMAND_UNKNOWN1 =  	bytes.fromhex('23000090')
RESPONSE_UNKNOWN1 = 	bytes.fromhex('28000091')

COMMAND_GET_HUB_INFO = 	bytes.fromhex('1e000090')
RESPONSE_GET_HUB_INFO = bytes.fromhex('4a000091')

COMMAND_MOVE_TO = 		bytes.fromhex('34000090')
REMPOSE_MOVE_TO = 		bytes.fromhex('34000091')

COMMAND_MOVE = 			bytes.fromhex('2d000090')

GET_ROOMS = 			bytes.fromhex('01000091')
GET_ROLLERS = 			bytes.fromhex('03000091')



"""Acmeda Pulse Hub Interface."""
import asyncio
import binascii
import logging

import async_timeout

import aiopulse.const as const
import aiopulse.utils as utils
import aiopulse.errors as errors
import aiopulse.elements as elements
import aiopulse.transport

_LOGGER = logging.getLogger(__name__)


class Hub:
    """Representation of an Acmeda Pulse Hub."""

    def __init__(self, host=None):
        """Init the hub."""
        self.topic = str.encode("Smart_Id1_y:")
        self.sequence = 4
        self.handshake = asyncio.Event()
        self.event_update = asyncio.Event()
        self.response_task = None
        self.running = False

        self.name = "Acmeda Pulse WiFi Hub"
        self.id = None
        self.host = host
        self.mac_address = None
        self.ip_address = None
        self.firmware_name = None
        self.wifi_module = None

        self.protocol = aiopulse.transport.HubTransportTcp(host)

        self.rollers = {}
        self.rooms = {}
        self.scenes = {}

        self.handshake.clear()

    @staticmethod
    async def discover(timeout=5):
        """Use a broadcast udp packet to find hubs on the lan."""
        discover_client = aiopulse.transport.HubTransportUdpBroadcast()

        await discover_client.connect()

        hubs = {}

        retries = 3

        try:
            with async_timeout.timeout(timeout * retries):
                for _ in range(1):
                    discover_client.send(const.HEADER + const.COMMAND_DISCOVER)
                    # discover_client.send(bytes.fromhex("000000030300001b"))
                    while True:
                        addr = None
                        try:
                            with async_timeout.timeout(timeout):
                                (_, addr) = await discover_client.receive()
                        except asyncio.TimeoutError:
                            pass

                        if addr and addr not in hubs:
                            _LOGGER.info("Discovered hub %s:%s", addr[0], addr[1])
                            try:
                                hub = Hub(addr[0])
                                await hub.connect()
                                await hub.disconnect()
                                hubs[addr] = hub
                                yield hub
                            except errors.NotConnectedException:
                                _LOGGER.error("Not connected")
        except asyncio.TimeoutError:
            pass
        _LOGGER.info("Discovery complete")

        await discover_client.close()

    async def connect(self, host=None):
        """Try and connect to the hub."""
        if host:
            self.host = host

        await self.protocol.connect(self.host)

        if self.handshake.is_set():
            _LOGGER.warn("Handshake already completed")
            return False

        if self.protocol.is_udp:  # udp
            # self.send_command(const.COMMAND_DISCOVER)
            # response = await self.get_response()
            self.send_command(const.COMMAND_CONNECT)
            raw_id = await self.get_response(const.RESPONSE_CONNECT)
        else:  # TCP
            self.send_command(const.COMMAND_CONNECT)
            raw_id = await self.get_response(const.RESPONSE_CONNECT)

        self.id = raw_id[2:].decode("utf-8")

        self.send_command(const.COMMAND_LOGIN + raw_id)
        response = await self.get_response(const.RESPONSE_LOGIN)

        if response[0] != 0:
            raise errors.InvalidResponseException

        self.send_command(
            const.COMMAND_SETID,
            bytes.fromhex("16000e0001000000000000000c000600120311073816ff9b"),
        )
        await self.get_response(const.RESPONSE_SETID)

        self.send_command(
            const.COMMAND_UNKNOWN1,
            bytes.fromhex("1100150002000000000000006002010030ffa9"),
        )
        await self.get_response(
            const.RESPONSE_UNKNOWN1
            + bytes.fromhex("06")
            + self.topic
            + bytes.fromhex("16000f0002000000000000000c000600120311073816ff9d")
        )
        await self.get_response(const.RESPONSE_SETID)

        _LOGGER.info("Handshake complete")
        self.handshake.set()

        return True

    async def disconnect(self):
        """Disconnect from the hub."""
        _LOGGER.debug("Disconnecting")
        await self.protocol.close()
        if not self.handshake.is_set():
            _LOGGER.warn("Not connected")
            return
        self.handshake.clear()
        _LOGGER.info("Disconnected")

    def send_command(self, command, message=None):
        """Send a command to the hub."""
        if message:
            self.protocol.send(
                const.HEADER + command + bytes.fromhex("05") + self.topic + message
            )
        else:
            self.protocol.send(const.HEADER + command)

    async def get_response(self, target_response=None):
        """Get a response, throw exception if it doesn't match expected response."""
        response = await self.protocol.receive()
        if not target_response:
            return response
        length = len(const.HEADER + target_response)
        if len(response) < length:
            raise errors.InvalidResponseException
        if response[0:length] != const.HEADER + target_response:
            raise errors.InvalidResponseException
        return response[length:]

    def response_hubinfo(self, message):
        """Receive start of hub information."""
        _LOGGER.debug("Received hub information")
        ptr = 10
        self.firmware_name, ptr = utils.unpack_string(message, ptr)
        ptr += 2
        _, ptr = utils.unpack_string(message, ptr)
        ptr += 2
        self.wifi_module, ptr = utils.unpack_string(message, ptr)
        ptr += 2
        self.mac_address, ptr = utils.unpack_string(message, ptr)
        ptr += 2
        self.ip_address, ptr = utils.unpack_string(message, ptr)

    def response_hubinfoend(self, message):
        """Receive end of hub information."""
        _LOGGER.debug("Received end of hub information")
        self.event_update.set()

    def response_roomlist(self, message):
        """Receive room list."""
        _LOGGER.debug("Received room list")
        ptr = 12
        room_count, ptr = utils.unpack_int(message, ptr, 1)
        for _ in range(room_count):
            _, ptr = utils.unpack_bytes(message, ptr, 2)
            room_id, ptr = utils.unpack_string(message, ptr)
            _, ptr = utils.unpack_bytes(message, ptr, 4)
            icon, ptr = utils.unpack_int(message, ptr, 1)
            _, ptr = utils.unpack_bytes(message, ptr, 2)
            room_name, ptr = utils.unpack_string(message, ptr)
            if room_id not in self.rooms:
                self.rooms[room_id] = elements.Room(self, room_id)
            self.rooms[room_id].icon = icon
            self.rooms[room_id].name = room_name

    def response_scenelist(self, message):
        """Receive scene list."""
        _LOGGER.debug("Received scene list")
        ptr = 12
        scene_count, ptr = utils.unpack_int(message, ptr, 1)
        ptr += 2
        for _ in range(scene_count):
            scene_id, ptr = utils.unpack_string(message, ptr)
            _, ptr = utils.unpack_bytes(message, ptr, 4)
            icon, ptr = utils.unpack_int(message, ptr, 1)
            _, ptr = utils.unpack_bytes(message, ptr, 2)
            scene_name, ptr = utils.unpack_string(message, ptr)
            _, ptr = utils.unpack_bytes(message, ptr, 7)
            _, ptr = utils.unpack_bytes(message, ptr)
            _, ptr = utils.unpack_bytes(message, ptr, 2)
            if scene_id not in self.scenes:
                self.scenes[scene_id] = elements.Scene(self, scene_id)
            self.scenes[scene_id].icon = icon
            self.scenes[scene_id].name = scene_name

    def response_timerlist(self, message):
        """Receive timer list."""
        _LOGGER.debug("Received timer list")
        pass

    def response_rollerlist(self, message):
        """Receive roller blind list."""
        _LOGGER.debug("Received roller list")
        ptr = 2  # sequence?
        ptr += 10
        roller_count, ptr = utils.unpack_int(message, ptr, 1)
        for _ in range(roller_count):
            ptr += 4  # unknown field
            roller_id, ptr = utils.unpack_int(message, ptr, 6)
            ptr += 2  # unknown field
            room_id, ptr = utils.unpack_string(message, ptr)
            ptr += 7  # unknown field
            roller_name, ptr = utils.unpack_string(message, ptr)
            ptr += 8  # unknown field
            roller_serial, ptr = utils.unpack_string(message, ptr)
            ptr += 4
            # battery level seems to be out of 25, convert to percentage
            roller_battery, ptr = utils.unpack_int(message, ptr, 1)
            roller_battery *= 4
            ptr += 5  # unknown field
            roller_percent, ptr = utils.unpack_int(message, ptr, 1)
            roller_flags, ptr = utils.unpack_int(message, ptr, 1)
            if roller_id not in self.rollers:
                self.rollers[roller_id] = elements.Roller(self, roller_id)
            self.rollers[roller_id].name = roller_name
            self.rollers[roller_id].serial = roller_serial
            self.rollers[roller_id].room_id = room_id
            if room_id in self.rooms:
                self.rollers[roller_id].room = self.rooms[room_id]
            else:
                self.rollers[roller_id].room = None
            self.rollers[roller_id].battery = roller_battery
            self.rollers[roller_id].closed_percent = roller_percent
            self.rollers[roller_id].flags = roller_flags
            self.rollers[roller_id].notify_callback()

    def response_authinfo(self, message):
        """Receive acmeda account information."""
        _LOGGER.debug("Received account information")
        ptr = 15
        _, ptr = utils.unpack_string(message, ptr)

    def response_position(self, message):
        """Receive change of roller position information."""
        _LOGGER.debug("Received change of roller position")
        ptr = 12
        roller_id, ptr = utils.unpack_int(message, ptr, 6)
        ptr += 10
        roller_percent, ptr = utils.unpack_int(message, ptr, 1)
        roller_flags, ptr = utils.unpack_int(message, ptr, 1)
        if roller_id in self.rollers:
            self.rollers[roller_id].closed_percent = roller_percent
            self.rollers[roller_id].flags = roller_flags
            self.rollers[roller_id].notify_callback()

    def response_discover(self, message):
        """Receive after discover broadcast packet."""
        _LOGGER.debug("Received broadcast response")
        pass

    class Receiver:
        """Wraps around a function that gets called for received messages."""

        def __init__(self, name, function):
            """Constructor for message receiver class."""
            self.name = name
            self.function = function

        def execute(self, target, message):
            """Executor function."""
            self.function(target, message)

    msgmap = {
        bytes.fromhex("1600"): Receiver("hub info", response_hubinfo),
        bytes.fromhex("4101"): Receiver("hub info end", response_hubinfoend),
        bytes.fromhex("0101"): Receiver("room list", response_roomlist),
        bytes.fromhex("3301"): Receiver("scene list", response_scenelist),
        bytes.fromhex("2101"): Receiver("roller list", response_rollerlist),
        bytes.fromhex("0800"): Receiver("auth info", response_authinfo),
        bytes.fromhex("2301"): Receiver("position", response_position),
        bytes.fromhex("2b01"): Receiver("position", response_position),
        bytes.fromhex("0f00"): Receiver("discover", response_discover),
    }

    def rec_ping(self, message):
        """Receive a ping from the hub."""
        _LOGGER.debug("Received hub ping response")
        pass

    def rec_message(self, message):
        """Receive and decode a message from the hub."""
        if message:
            if message[0] != 6:
                raise errors.InvalidResponseException

            if message[1 : (1 + len(self.topic))] != self.topic:
                raise errors.InvalidResponseException
            ptr = 1 + len(self.topic)
            _, ptr = utils.unpack_int(message, 2, ptr)
            mtype = message[ptr : (ptr + 2)]
            ptr = ptr + 2
            if mtype in self.msgmap:
                _LOGGER.info("Parsing %s", self.msgmap[mtype].name)
                self.msgmap[mtype].execute(self, message[ptr:])
            else:
                _LOGGER.warning(
                    "Unable to parse message %s message %s",
                    binascii.hexlify(mtype),
                    binascii.hexlify(message),
                )

    respmap = {
        const.RESPONSE_PING: Receiver("ping", rec_ping),
        bytes.fromhex("03000091"): Receiver("acknowledge", rec_message),
        bytes.fromhex("01000091"): Receiver("scene list", rec_message),
        bytes.fromhex("02000091"): Receiver("roller list", rec_message),
        bytes.fromhex("23000091"): Receiver("hub info end", rec_message),
        bytes.fromhex("28000091"): Receiver("discover", rec_message),
        bytes.fromhex("34000091"): Receiver("moving", rec_message),
        bytes.fromhex("42000091"): Receiver("unknown", rec_message),
        bytes.fromhex("43000091"): Receiver("account info", rec_message),
        bytes.fromhex("44000091"): Receiver("response move", rec_message),
        const.RESPONSE_DISCOVER: Receiver("discover", rec_message),
        bytes.fromhex("5b000091"): Receiver(
            "room list", rec_message
        ),  # after adding second room
        bytes.fromhex("5F000091"): Receiver("hub info", rec_message),
        bytes.fromhex("60000091"): Receiver("room list", rec_message),
        bytes.fromhex("65000091"): Receiver("timer list", rec_message),
        bytes.fromhex("62000091"): Receiver(
            "62000091", rec_message
        ),  # Unknown from quentinsf
        bytes.fromhex("04000091"): Receiver(
            "04000091", rec_message
        ),  # Unknown from quentinsf
    }

    def response_parse(self, response):
        """Decode response."""
        if response[0:4] != bytes.fromhex("00000003"):
            _LOGGER.warning("Unknown response: %s", binascii.hexlify(response[0:4]))
            raise errors.InvalidResponseException

        message_type = response[4:8]

        if message_type[0] > 127:
            message_type = response[5:9]
            message = response[9:]
        else:
            message = response[8:]

        mtype = bytes(message_type)
        _LOGGER.debug(
            "Received message type: %s message: %s",
            binascii.hexlify(mtype),
            binascii.hexlify(message),
        )
        if mtype in Hub.respmap:
            _LOGGER.info("Received %s", Hub.respmap[mtype].name)
            Hub.respmap[mtype].execute(self, message)
        else:
            _LOGGER.warning(
                "Received unknown message type: %s message: %s",
                binascii.hexlify(mtype),
                binascii.hexlify(message),
            )
            self.rec_message(message)

    async def response_parser(self):
        """Receive a response from the hub and work out what message it is."""
        try:
            _LOGGER.info("Starting response parser")
            while True:
                try:
                    with async_timeout.timeout(30):
                        response = await self.protocol.receive()
                    if len(response) > 0:
                        self.response_parse(response)
                except asyncio.TimeoutError:
                    self.send_command(const.COMMAND_PING)
                except errors.InvalidResponseException:
                    self.send_command(const.COMMAND_PING)
        except errors.NotConnectedException:
            _LOGGER.info("Disconnected, stopping parser")

    async def update(self):
        """Update all hub information (includes scenes, rooms, and rollers)."""
        self.event_update.clear()
        await self.send_payload(
            const.COMMAND_GET_HUB_INFO,
            bytes.fromhex("F000"),
            bytes.fromhex("000000000000FF"),
        )
        _LOGGER.info("Hub update command sent")

    async def send_payload(self, command, message_type, message):
        """Send payload to the hub."""
        if not self.running:
            raise errors.NotRunningException
        await self.handshake.wait()
        data = message_type + utils.pack_int(self.sequence, 2) + message
        checksum = bytes([sum(data) & 0xFF])
        self.sequence += 2
        command_header = const.HEADER + command + bytes.fromhex("05") + self.topic
        length = len(data) + 1  # bytes.fromhex('0C00')
        self.protocol.send(command_header + utils.pack_int(length, 2) + data + checksum)

    async def run(self):
        """Start hub by connecting then awaiting for messages.

        Runs until the stop() method is called.
        """
        if self.running:
            _LOGGER.warn("Already running")
            return
        self.running = True
        while self.running:
            try:
                _LOGGER.info("Connecting")
                await self.connect()
            except errors.InvalidResponseException:
                _LOGGER.warn("Connect failed")
                continue
            try:
                await self.update()
                await self.response_parser()
            except errors.InvalidResponseException:
                _LOGGER.warn("Connect failed")
            if self.running:
                await self.disconnect()
                await asyncio.sleep(5)
        _LOGGER.debug("Stopped")

    async def stop(self):
        """Tell hub to stop and await for it to disconnect."""
        if not self.running:
            _LOGGER.warn("Already stopped")
            return
        _LOGGER.debug("Stopping")
        self.running = False
        await self.disconnect()



"""Pip setup file for aiopulse library."""
from distutils.core import setup

setup(
    name="aiopulse",
    packages=["aiopulse"],
    version="0.2.2",
    license="apache-2.0",
    description="Python module for Rollease Acmeda Automate integration.",
    author="Alan Murray",
    author_email="pypi@atmurray.net",
    url="https://github.com/atmurray/aiopulse",
    download_url="https://github.com/atmurray/aiopulse/archive/v0.2.2.tar.gz",
    keywords=["automation"],
    install_requires=["asyncio", "async_timeout"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
    ],
)



"""Network transport abstraction for hub."""
import logging
import asyncio
import socket

from aiopulse.errors import NotConnectedException

_LOGGER = logging.getLogger(__name__)


class HubTransportBase(asyncio.Protocol):
    """Base class for Hub transport implementations."""

    def __init__(self):
        """Constructor for the base transport class."""
        self.transport = None

    def connection_made(self, transport):
        """Called when a connection is made."""
        _LOGGER.debug("Connection established")
        self.transport = transport

    def error_received(self, exc):
        """Called when an error is received."""
        _LOGGER.error("Error received: %s", exc)

    def connection_lost(self, exc):
        """Called when a connection is lost."""
        _LOGGER.info("Socket closed")


class HubTransportUdp(HubTransportBase):
    """UDP Based Hub transport."""

    def __init__(self, host=None, port=12414):
        """Constructor for UDP transport class."""
        self.host = host
        self.port = port
        self.transport = None
        self.protocol = None
        self.is_udp = True
        self.receive_queue = asyncio.Queue()
        super().__init__()

    async def connect(self, host=None):
        """Initialise connection."""
        if host:
            self.host = host

        loop = asyncio.get_event_loop()
        self.transport, self.protocol = await loop.create_datagram_endpoint(
            lambda: self, remote_addr=(self.host, self.port),
        )

    async def close(self):
        """Close the connection."""
        self.transport.close()
        _LOGGER.info("UDP connection closed")

    def send(self, buffer):
        """Abstraction of the underlying transport to send a buffer."""
        self.transport.sendto(buffer, (self.host, self.port))

    async def receive(self):
        """Abstraction of the underlying transport to receive."""
        return await self.receive_queue.get()

    def datagram_received(self, data, addr):
        """Callback for a received datagram, enqueue it."""
        # Don't close the socket as we might get multiple responses.
        _LOGGER.debug("UDP datagram received")
        self.receive_queue.put_nowait((data, addr))


class HubTransportUdpBroadcast(HubTransportUdp):
    """UDP Based Hub transport."""

    async def connect(self, host="255.255.255.255"):
        """Init connection."""
        if host:
            self.host = host
        addrinfo = socket.getaddrinfo(self.host, None)[0]
        sock = socket.socket(addrinfo[0], socket.SOCK_DGRAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)

        loop = asyncio.get_event_loop()
        self.transport, self.protocol = await loop.create_datagram_endpoint(
            lambda: self, sock=sock,
        )


class HubTransportTcp(HubTransportBase):
    """TCP based Hub transport."""

    def __init__(self, host=None):
        """TCP Transport constructor."""
        self.host = host
        self.port = 12416

        self.reader = None
        self.writer = None
        self.transport = None
        self.protocol = None
        self.is_udp = False
        super().__init__()

    async def connect(self, host=None):
        """Init connection."""
        if host:
            self.host = host

        if self.writer:
            _LOGGER.warn("Already connected.")
            return

        loop = asyncio.get_event_loop()
        self.reader = asyncio.StreamReader(loop=loop)
        # super().__init__(self.reader)
        self.protocol = asyncio.StreamReaderProtocol(self.reader, loop=loop)
        self.transport, _ = await loop.create_connection(
            lambda: self, self.host, self.port
        )
        self.writer = asyncio.StreamWriter(
            self.transport, self.protocol, self.reader, loop
        )

    async def close(self):
        """Close the connection."""
        if not self.writer:
            _LOGGER.error("Not connected")

        self.writer.close()
        await self.writer.wait_closed()
        self.writer = None
        _LOGGER.info("TCP connection closed.")

    def send(self, buffer):
        """Abstraction of the underlying transport to send a buffer."""
        self.writer.write(buffer)

    async def receive(self):
        """Receive from stream."""
        if self.writer.is_closing():
            raise NotConnectedException
        return await self.reader.read(65535)

    def data_received(self, data):
        """Callback when data has been received."""
        _LOGGER.debug("TCP data received.")
        self.protocol.data_received(data)

    def connection_made(self, transport):
        """Callback when a connection has been made."""
        self.protocol.connection_made(transport)
        super().connection_made(transport)

    def connection_lost(self, exc):
        """Callback when a connection is lost."""
        self.protocol.connection_lost(exc)
        super().connection_lost(exc)
